﻿using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// 
  /// </summary>
  [Item("ParameterConfigurationCrossover", "TODO")]
  [StorableClass]
  public class ParameterConfigurationCrossover : SingleSuccessorOperator, IParameterConfigurationOperator, ICrossover {
    public override bool CanChangeName {
      get { return false; }
    }

    public ILookupParameter<IRandom> RandomParameter {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ILookupParameter<ItemArray<IValueConfiguration>> ParentsParameter {
      get { return (ScopeTreeLookupParameter<IValueConfiguration>)Parameters["Parents"]; }
    }
    public ILookupParameter<IValueConfiguration> ChildParameter {
      get { return (ILookupParameter<IValueConfiguration>)Parameters["Child"]; }
    }

    public IValueLookupParameter<IIntValueCrossover> IntValueCrossoverParameter {
      get { return (IValueLookupParameter<IIntValueCrossover>)Parameters[MetaOptimizationProblem.IntValueCrossoverParameterName]; }
    }
    public IValueLookupParameter<IDoubleValueCrossover> DoubleValueCrossoverParameter {
      get { return (IValueLookupParameter<IDoubleValueCrossover>)Parameters[MetaOptimizationProblem.DoubleValueCrossoverParameterName]; }
    }

    [StorableConstructor]
    protected ParameterConfigurationCrossover(bool deserializing) : base(deserializing) { }
    protected ParameterConfigurationCrossover(ParameterConfigurationCrossover original, Cloner cloner) : base(original, cloner) { }
    public ParameterConfigurationCrossover()
      : base() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator which should be used for stochastic crossover operators."));
      Parameters.Add(new ScopeTreeLookupParameter<IValueConfiguration>("Parents", "The parent vectors which should be crossed."));
      ParentsParameter.ActualName = "ParameterConfigurationTree";
      Parameters.Add(new LookupParameter<IValueConfiguration>("Child", "The child vector resulting from the crossover."));
      ChildParameter.ActualName = "ParameterConfigurationTree";

      Parameters.Add(new ValueLookupParameter<IIntValueCrossover>(MetaOptimizationProblem.IntValueCrossoverParameterName, ""));
      Parameters.Add(new ValueLookupParameter<IDoubleValueCrossover>(MetaOptimizationProblem.DoubleValueCrossoverParameterName, ""));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ParameterConfigurationCrossover(this, cloner);
    }

    public override IOperation Apply() {
      IValueConfiguration child1 = (IValueConfiguration)ParentsParameter.ActualValue[0].Clone();
      IValueConfiguration child2 = (IValueConfiguration)ParentsParameter.ActualValue[1];

      //child1.Cross(child2, RandomParameter.ActualValue);
      //this.ChildParameter.ActualValue = child1;

      child1.Cross(RandomParameter.ActualValue, child2, Cross, this);
      this.ChildParameter.ActualValue = child1;

      return base.Apply();
    }

    private static void Cross(IRandom random, IOptimizable configuartion, ParameterConfigurationCrossover pccross) {
      if (configuartion is IValueConfiguration) {
        var vc = configuartion as IValueConfiguration;
        var value = vc.ActualValue.Value;
        var range = vc.RangeConstraint;

        if (value is IntValue) {
          pccross.IntValueCrossoverParameter.ActualValue.Apply(random, (IntValue)value, (IntValueRange)range);
        } else if (value is PercentValue) {
          pccross.DoubleValueCrossoverParameter.ActualValue.Apply(random, (PercentValue)value, ((PercentValueRange)range).AsDoubleValueRange());
        } else if (value is DoubleValue) {
          pccross.DoubleValueCrossoverParameter.ActualValue.Apply(random, (DoubleValue)value, (DoubleValueRange)range);
        }
      } else if (configuartion is IParameterConfiguration) {

      }
    }

    private IntValue CrossInteger(IParameterConfiguration parameter1, IParameterConfiguration parameter2) {
      IntegerVector integerChild = HeuristicLab.Encodings.IntegerVectorEncoding.DiscreteCrossover.Apply(
        RandomParameter.ActualValue,
        new IntegerVector(new IntArray(new int[] { ((IntValue)parameter1.ActualValue.Value).Value })),
        new IntegerVector(new IntArray(new int[] { ((IntValue)parameter2.ActualValue.Value).Value })));
      return new IntValue(integerChild[0]);
    }

  }
}
